from abc import ABC, abstractmethod

from .service_types import ServiceTypes


class ActionSystemBase(ABC):
    def __init__(self, client):
        self.client = client

    @abstractmethod
    def trigger_action(self, service_type: ServiceTypes, slug: str, action_name: str, parameters: dict) -> dict:
        self.client.validate_service_access(service_type, slug)



from .action_system import BoursikaActionSystem
from .event_system import BoursikaEventSystem
from .predictor_system import BoursikaPredictorSystem
from ..core import StrategyClientBase, PredictorClientBase


class BoursikaStrategyClient(StrategyClientBase):
    _action_system_class = BoursikaActionSystem
    _event_system_class = BoursikaEventSystem
    _predictor_system_class = BoursikaPredictorSystem


class BoursikaPredictorClient(PredictorClientBase):
    _action_system_class = BoursikaActionSystem
    _event_system_class = BoursikaEventSystem
    _predictor_system_class = BoursikaPredictorSystem



from .action_system import ActionSystemBase
from .client import StrategyClientBase, PredictorClientBase
from .error import ActionError
from .event_system import EventSystemBase
from .predictor_system import PredictorSystemBase
from .service_types import ServiceTypes



import json
import logging
from functools import wraps

from websocket import WebSocketApp

from ..config import BOURSIKA_WS_BASE_URL
from ..core import EventSystemBase, ServiceTypes

logger = logging.getLogger(__name__)


class BoursikaEventSystem(EventSystemBase):
    def start(self) -> None:
        ws = WebSocketApp(f'{BOURSIKA_WS_BASE_URL}/ws/event-system/',
                          on_open=_method_to_function(self, self._on_open),
                          on_message=_method_to_function(self, self._on_message),
                          on_error=_method_to_function(self, self._on_error),
                          on_close=_method_to_function(self, self._on_close))
        ws.run_forever()
        pass

    def _on_open(self, ws: WebSocketApp):
        try:
            body = {
                'action': 'listen',
                'payload': {
                    'events': []
                }
            }

            for event_descriptor in self._event_listener_descriptors:
                body['payload']['events'].append({
                    'service_type': event_descriptor.service_type.value,
                    'slug': event_descriptor.slug,
                    'event_name': event_descriptor.event_name,
                })

            ws.send(json.dumps(body))
        except Exception as ex:
            logger.exception(ex)
            raise ex

    def _on_message(self, ws: WebSocketApp, message: str):
        try:
            content = json.loads(message)

            if content.get('message_type') == 'EVENT':
                service_type = ServiceTypes(content.get('service_type'))
                slug = content.get('slug')
                event_name = content.get('event_name')
                payload = content.get('payload')

                super().trigger_event(service_type=service_type, slug=slug, event_name=event_name, payload=payload)
        except Exception as ex:
            logger.exception(ex)

    def _on_error(self, ws: WebSocketApp, error: Exception):
        logger.exception(error)

    def _on_close(self, ws: WebSocketApp):
        pass


def _method_to_function(container, callback):
    @wraps(callback)
    def wrapper(*args, **kwargs):
        return callback(*args, **kwargs)

    return wrapper



import requests

from ..config import BOURSIKA_API_BASE_URL
from ..core import ActionSystemBase, ServiceTypes, ActionError


class BoursikaActionSystem(ActionSystemBase):
    def trigger_action(self, service_type: ServiceTypes, slug: str, action_name: str, parameters: dict) -> dict:
        super().trigger_action(service_type, slug, action_name, parameters)

        if service_type == ServiceTypes.PROVIDER:
            endpoint = f'{BOURSIKA_API_BASE_URL}/providers/{slug}/actions/{action_name}/calls/'
        elif service_type == ServiceTypes.TRADER:
            endpoint = f'{BOURSIKA_API_BASE_URL}/traders/{slug}/actions/{action_name}/calls/'
        else:
            raise Exception('this type of service doesn\'t support actions')

        try:
            response = requests.post(endpoint, json={
                'parameters': parameters,
            }, headers={
                'ApiKey': self.client.api_key,
            }, timeout=10)

            response.raise_for_status()

            return response.json().get('result')
        except requests.exceptions.HTTPError as e:
            raise ActionError(message=e.response.json().get('message'))



from typing import Callable

from .action_system import ActionSystemBase
from .service_types import ServiceTypes


class ServiceActionProxy:
    def __init__(self, action_system: ActionSystemBase, service_type: ServiceTypes, slug: str):
        self.action_system = action_system
        self.service_type = service_type
        self.slug = slug

    def __getattr__(self, action_name: str) -> Callable:
        def f(**kwargs: dict):
            return self.action_system.trigger_action(service_type=self.service_type,
                                                     slug=self.slug,
                                                     action_name=action_name,
                                                     parameters=kwargs)

        return f


class ServiceActionProxyFactory:
    def __init__(self, action_system: ActionSystemBase, service_type: ServiceTypes):
        self.action_system = action_system
        self.service_type = service_type

    def __getattr__(self, slug: str) -> ServiceActionProxy:
        return ServiceActionProxy(action_system=self.action_system, service_type=self.service_type, slug=slug)
